package abstractPracticeTheory;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import repast.simphony.context.Context;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.parameter.Parameters;
import repast.simphony.query.space.grid.GridCell;
import repast.simphony.query.space.grid.GridCellNgh;
import repast.simphony.query.space.grid.MooreQuery;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.grid.Grid;
import repast.simphony.space.grid.GridPoint;
import repast.simphony.util.ContextUtils;
import repast.simphony.util.SimUtilities;

public class Individual {
	
	private Grid<Object> grid;
	
	//Variable to store bestPractice in between call of step() and postStep()
	protected Material bestMaterial;
	
	protected Meaning meaning;
	protected Material material;
	protected Competence competence;
	
	protected boolean deliberateChoice;
	protected boolean moved;
	protected double habitStrength;
	
	public Individual(){
		deliberateChoice = true;
		moved = false;
		habitStrength=0.0;
	}
	
	public Individual(Meaning m, Competence s, Material mat){
		this();
		meaning = m;
		competence = s;
		material = mat;
	}
	
	public Individual(Meaning m, Competence s, Material mat, Grid<Object> g){
		
		this(m,s,mat);
		setGrid(g);
	}
	
	/**Create meaning and skill of length nrBits with random bits.*/
	public void setUpRandom(int nrBits){
		
		meaning = new Meaning(nrBits);
		meaning.setUpRandom();
		competence = new Competence(nrBits);
		competence.setUpRandom();
	}
	
	public Grid getGrid(){
		return grid;
	}
	
	public void setGrid(Grid<Object> g){
		grid = g;
	}
	
	//set discursiveConsciousness to false
	//(for all Individuals before step is called for any Individual)
	@ScheduledMethod(start=1, interval=1, shuffle=true, priority=6.0)  
	public void preStep(){
		deliberateChoice = false;
		moved = false;
	}
	
	//step has higher priority than postStep in order to be executed earlier
	//(for all Individuals before postStep is called for any Individual)
	@ScheduledMethod(start=1, interval=1, shuffle=true, priority=2.0)  
	public void step(){
		//note: asynchronic update of meaning and skills: if there are several processes to change meaning (e.g. adaptation to material
		//and normative influence contradictory influences may occur which are difficult to implement in synchronic update
		
		//check which processes are set active in the GUI and execute them
		Parameters params = RunEnvironment.getInstance().getParameters();
		
//		if((Boolean) params.getValue("adaptMeaning")){	
			adaptMeaning();
//		}
//		if((Boolean) params.getValue("adaptSkills")){	
			adaptCompetence();
//		}
		
		bestMaterial=null;				//refresh to indicate whether a new Material has been chosen or not
		if(!deliberateChoice){	//may be activated externally
			deliberateChoice = endogenousDeliberateChoice();
		}
				
		//if active search, move is active and there is some space on the grid
		/*if(discursiveConsciousness && ((Integer) params.getValue("move") > 0) && (Integer)(params.getValue("percentOccupiedCells"))<100){
			move();
		}*/
		
		socialInfluence((Integer) params.getValue("socialInfluence"));
		
		if(deliberateChoice){
			//socialInfluence((Integer) params.getValue("socialInfluence")); //uncomment this line and comment same line above to model that exchange happens only when there is active search for alternatives.
			chooseMaterial();
		}
		
		habituation((Double) params.getValue("ha"));
	}
	
	//replace the Practice on this GridCell with bestPractice if a new practice has been chosen
	//clear memory of innovation if not chosen
	@ScheduledMethod(start=1, interval=1, shuffle=true, priority=1.0)
	public void postStep(){
		
		if(bestMaterial != null){	//a new Material has been chosen
			Context<Object> context = ContextUtils.getContext(material);
			context.remove(material); //delete current practice
			Material nM = bestMaterial.cloneThis(); //copy bestPractice
			context.add(nM);
			material = nM;
			((ColoredMaterial)material).setGrid(grid);
			//move it to this individuals location
			GridPoint pt = grid.getLocation(this);
			grid.moveTo(nM,pt.getX(),pt.getY());	
		}
	}
	
	/**Selects the material with the highest coherence with this individual's competence and meaning.
	 * Material are checked which are available in the (moore) neighbourhood 
	 */
	protected void chooseMaterial(){
		
		//get all material in neighbourhood
		MooreQuery<Object> query = new MooreQuery<Object>(grid, this);
		List<ColoredMaterial> availableMaterials = new ArrayList();
		for(Object o: query.query()){
			if(o instanceof ColoredMaterial){
				availableMaterials.add((ColoredMaterial)o);
			}
		}	
		
		//assume the current practice is the best one
		double max = coherence();
		ColoredMaterial m;
		if(availableMaterials.size()>0){
			for(int i=0;i<availableMaterials.size();i++){
				m = availableMaterials.get(i);
				if(coherence(m) > max){				//check whether this material fits this individual even better than current best
					bestMaterial = m;						//if so take this as new best material
					max = coherence(m);
				}
			}
		}
		
	}
	
	
	/*
	 * Competence und Meaning werden separat mit der Material verglichen. Die schlechtere Kohrenz
	 * wird zurckgegeben.*/
/**/public double coherence(Material m){
		
		double sumCompetence =0.0;
		double sumMeaning = 0.0;
		
		for(int i=0; i<m.length(); i++){
			sumCompetence+= delta(competence.get(i), m.get(i));
			sumMeaning+= delta(meaning.get(i), m.get(i));
		}
		return sumCompetence < sumMeaning ? sumCompetence/m.length() : sumMeaning/m.length();
	}
 /**/	
	
	public double coherence(){
		
		if(this.material != null){
			return coherence(this.material);
		}
		return -1.0;
	}
	
	/*Returns 1 if a == b, 0 otherwise.*/
	private int delta(int a, int b){
		if(a==b)return 1; 
		return 0;
	}
	
	protected void habituation(double ha){
		
		if(bestMaterial != null){	//a new material has been chosen
			habitStrength = 0.0;
		}
		else{
			habitStrength = habitStrength + (1-habitStrength)*ha;
		}
	}
	
	
	
	/**Returns true if this Individual acts routinely 
	 * (discursive consciousness is set false in preStep() and then potentially activated later on)
	 * */
	public boolean routineBehaviour(){
		return !this.deliberateChoice;
	}
	
	public Material getMaterial(){
		return this.material;
	}
	
	public void offerMaterial(Material m){
		//to be implemented
	}
	
		
	protected boolean endogenousDeliberateChoice(){
		
		Parameters params = RunEnvironment.getInstance().getParameters();
		//habiutal behaviour is more likely the more the stronger the habit is, and the higher the coherence of the individuals current practice is
		//Consequently, discursive consciousness is activated if random disturbance (prob depends on incoherence) is stronger than habit
		if((RandomHelper.nextDoubleFromTo(Math.min(0.5,1-coherence()), 0.501) > 0.5*habitStrength)){	
			return true; 	
		}
		return false;
	}
	public void setDeliberateChoice(boolean b){
		deliberateChoice = b;
	}
	
	public boolean getDeliberateChoice(){
		return deliberateChoice;
	}
	
	/**Meaning is adapted towards the material component, i.e. cognitive dissonance is reduced. 
	 * Select one bit randomly. If there is a difference between meaning and material then adapt the meaning 
	 * bit to that of the material. 
	 */
	protected void adaptMeaning(){
		
		int pos = RandomHelper.nextIntFromTo(0, meaning.length()-1); 
		
		//take current material because adaptMeaning is executed before choosePractice()
		if(material.get(pos) != meaning.get(pos)){
				meaning.set(pos, material.get(pos));
		}	
	}
	/** Gain skills which make this practitioner fitter: select one bit randomly. Change it to value of material bit, if different.*/ 
	protected void adaptCompetence(){
		int pos = RandomHelper.nextIntFromTo(0, competence.length()-1);
		
		//take current practice p because adaptSkills is executed before choosePractice()
		if(material.get(pos) != competence.get(pos)){
				competence.set(pos, material.get(pos));
		}
	}
	/**
	 * exchange meaning with one neighbour per time step on random locus 
	 * The probability that exchange actually occurs depends on similarity between individuals
	 * Probability is linear with similarity -> only individuals with strictly opposed meaning will
	 * never exchange.
	 * If exchange occurs and value at locus is different, adapt own locus. 
	 * This transfers social pressure to the agent. (cf. Axelrod model on dissemination of culture)
	 */
	protected void socialInfluence(int nrLoci){
		
		//get all neighbours
		MooreQuery<Object> query = new MooreQuery<Object>(grid, this);
		List<Individual> neighbours = new ArrayList();
		for(Object o: query.query()){
			if(o instanceof Individual){
				neighbours.add((Individual)o);
			}
		}	
		
		if(neighbours.size()>0){ //if there are any neighbours
			//choose random neighbour
			Individual neighbour = neighbours.get(RandomHelper.nextIntFromTo(0, neighbours.size()-1)); 
	
			//calculate probability for exchange (linearly dependent on similarity)
			//opposing values '0'<->'1' are counted as not similar
			double similarLoci=0;
			for(int i=0;i<this.meaning.length();i++){
				if(this.meaning.get(i)==neighbour.meaning.get(i))
					similarLoci++;	
			}
			double pExchange = similarLoci/this.meaning.length();
			
			//if exchange happens, adapt random nrLoci own loci to neighbours value
			//if they are already similar nothing happens
			if(RandomHelper.nextDoubleFromTo(0, 1)<pExchange){
				for(int i=0; i<nrLoci;i++){
					int randomLocus = RandomHelper.nextIntFromTo(0, meaning.length()-1); 
					this.meaning.set(randomLocus, neighbour.meaning.get(randomLocus));
				}
			}
		}
	}
	
	/**If unhappy with neighbours move to random empty cell.*/
	protected void move(){
					
		//check similarity of neighbourhood
		//get all neighbours
		MooreQuery<Object> query = new MooreQuery<Object>(grid, this);
		List<Individual> neighbours = new ArrayList();
		for(Object o: query.query()){
			if(o instanceof Individual){
				neighbours.add((Individual)o);
			}
		}	
		
		double highestSimilarity =0.0;
		if(neighbours.size()>0){ //if there are any neighbours
				//check all neighbours
				for(int n=0; n< neighbours.size();n++){
					Individual neighbour = neighbours.get(n);	
					//calculate similarity
					double similarLoci=0;
					for(int i=0;i<this.meaning.length();i++){
						if(this.meaning.get(i)==neighbour.meaning.get(i))
							similarLoci++;	
					}
					double similarity = similarLoci/this.meaning.length();
					if(similarity > highestSimilarity){
						highestSimilarity = similarity;
					}
				}
		}

		
		//with probability anti-proportional to max similarity to neighbour move to random new location
		if(RandomHelper.nextDouble()<1.0-highestSimilarity){
			
			int width  = grid.getDimensions().getWidth();
			int height = grid.getDimensions().getHeight();
	
			//choose random location
			int x = RandomHelper.nextIntFromTo(0, width-1);
			int y = RandomHelper.nextIntFromTo(0, height-1);
			
			//if occupied, search new one
			while(grid.getObjectAt(x,y) != null){
				x = RandomHelper.nextIntFromTo(0, width-1);
				y = RandomHelper.nextIntFromTo(0, height-1);
			}
	
			//move Individual and associated Material to new location
			grid.moveTo(this, x,y);
			grid.moveTo(material,x,y);
			moved = true;
		}
	}
	
	public boolean getMoved(){
		return moved;
	}
	
	public double getHabitStrength(){
		return habitStrength;
	}
}
